//*----------------------------------------------------------------------------
//*         ATMEL Microcontroller Software Support  -  ROUSSET  -
//*----------------------------------------------------------------------------
//* The software is delivered "AS IS" without warranty or condition of any
//* kind, either express, implied or statutory. This includes without
//* limitation any warranty or condition with respect to merchantability or
//* fitness for any particular purpose, or against the infringements of
//* intellectual property rights of others.
//*----------------------------------------------------------------------------
//* File Name           : main.c
//* Object              : main application written in C
//* Creation            : Hi   11/18/2002
//*
//*----------------------------------------------------------------------------
#include "AT91RM9200.h"
#ifdef __GNUC__
#include "proto.h"			
#else 
#include "lib_AT91RM9200.h"
#endif							/* __GNUC__ */
#include "emac.h"

#define AT91C_DISPLAY_ALL_IPHEADER	1	// 
#define RTL8201BL 						// RTL8201BL
#undef  PHY_RMII						// MII interface
#define PHY_ADDR 	(0x1F ^ 2)			// PHY address

#define SDRAM_ADDR   		        (0x20000000)  // SDRAM bank address

extern unsigned int AT91F_GetTickCount(void);
extern void AT91F_DBGU_Printk(char *buffer);
extern void AT91F_DBGU_Printx(unsigned long val);
extern void AT91F_DBGU_Printl(unsigned long val);
extern void AT91F_DBGU_Printv(char * pref, unsigned long val, char *suf);
extern void AT91F_DBGU_Printw(char * pref, unsigned long val, char *suf);

// Our Ethernet MAC address and IP address
char OurEmacAddr[6] = {0x01, 0x02, 0x03, 0x04, 0x05, 0x06};
unsigned char OurIpAddr[4]  = {10, 0, 0, 2}; // {0x0A, 0xD7, 0xF5, 0xAC}

//Buffer descriptor address must be word aligned
#define AT91C_EMAC_TDLIST_BASE	0x20010000

//char RxPacket[NB_ETH_RX_PACKETS*ETH_PACKET_SIZE];
char * RxPacket = (char *) SDRAM_ADDR;
AT91PS_TdDescriptor tdList = (AT91PS_TdDescriptor)AT91C_EMAC_TDLIST_BASE;



// ****************************************************************************************************
// ** EMAC lowlevel functions 
// ****************************************************************************************************
//*----------------------------------------------------------------------------
//* \fn    AT91F_MII_ReadPhy
//* \brief This function return the MII phy register
//*----------------------------------------------------------------------------
int AT91F_MII_ReadPhy(
	AT91PS_EMAC    pEmac,     // \arg Pointer to AT91PS_EMAC service
	unsigned char   addr,
	unsigned short * val
	) 
{
	unsigned int Wait = AT91F_GetTickCount() + 100;
	unsigned int value = 0x60020000 | (addr << 18) | (PHY_ADDR << 23);
/* 29:28 10b - read PHY register, 01b write PHY register */
/* 30 - must be 1 */
/* 31 - must be 0 */
	pEmac->EMAC_MAN = value;
  	while(AT91F_GetTickCount() < Wait) {
		if((pEmac->EMAC_SR & AT91C_EMAC_IDLE) != 0) {
		/* 0 - PHY logic IDLE, 1 - PHY running */
			if(val != 0) {
				*val = (pEmac->EMAC_MAN & 0x0000ffff);
				return(1);
			}
		}
 
  	}

	if(val != 0) {
		*val = 0;
	}
	return (0);
}

int AT91F_MII_WritePhy(AT91PS_EMAC pEmac, unsigned char addr, unsigned short val) 
{
	unsigned int Wait = AT91F_GetTickCount() + 100;
	unsigned int value = 0x50020000 | (addr << 18) | (PHY_ADDR << 23) |
		(val & 0xFFFF);
/* 29:28 10b - read PHY register, 01b write PHY register */
/* 30 - must be 1 */
/* 31 - must be 0 */
	pEmac->EMAC_MAN = value;
  	while(AT91F_GetTickCount() < Wait) {
		if((pEmac->EMAC_SR & AT91C_EMAC_IDLE) != 0) {
		/* 0 - PHY logic IDLE, 1 - PHY running */
			return(1);
		}
 
  	}
	return (0);
}



//*----------------------------------------------------------------------------
//* \fn    AT91F_MII_GetLinkSpeed
//* \brief This function configure the link speed
//*----------------------------------------------------------------------------
int MII_GetLinkSpeed(AT91PS_EMAC pEmac)     // \arg Pointer to AT91PS_EMAC service
{
#ifdef RTL8201BL 
  	unsigned short stat0; 				// PHY register 0
  	unsigned short stat1;				// PHY register 1
  	unsigned short stat25; 				// PHY test register 25

	AT91F_DBGU_Printk("Reset PHY\n");
 	if(AT91F_MII_WritePhy(pEmac, 0, 0x8000 ) == 0) {
		AT91F_DBGU_Printk("Reset PHY error\n");
 	}

	AT91F_DBGU_Printk("Restart autoneg\n");
 	if(AT91F_MII_WritePhy(pEmac, 0, 0x1200 ) == 0) {
		AT91F_DBGU_Printk("Restart autoneg. error\n");
 	}

	while(1) { 
		AT91F_MII_ReadPhy(pEmac, 1, & stat1);
		if(AT91F_MII_ReadPhy(pEmac, 1, & stat1) == 0) {
			AT91F_DBGU_Printk("PHY status1 register read error\n");
		}

		if( (stat1 & (1 << 5)) != 0) {
   			AT91F_DBGU_Printk("AN comp\n");
			break;
		}
	}

	AT91F_DBGU_Printw("PHY status register", stat1, "\n");

	AT91F_MII_ReadPhy(pEmac,  0, & stat0);		/* read status 0 register */ 
	AT91F_MII_ReadPhy(pEmac, 25, & stat25);		/* read test register */ 
   	AT91F_DBGU_Printw("PHY 0 register" , stat0 , "\n");
   	AT91F_DBGU_Printw("PHY 25 register", stat25, "\n");
	if( (stat0 & (1 << 8)) != 0 ) {
	/* full duplex */
		pEmac->EMAC_CFG |= AT91C_EMAC_FD;
 		if( (stat25 & 1) != 0) {
			/* full duplex 100Mb */		
   			AT91F_DBGU_Printk("100Mb Full dup\n");
    		pEmac->EMAC_CFG |=   AT91C_EMAC_SPD;
		} else {
			/* full duplex 10Mb */
   			AT91F_DBGU_Printk("10Mb Full dup\n");
    		pEmac->EMAC_CFG &= (~AT91C_EMAC_SPD);
		}
	} else {
	/* half duplex */
		pEmac->EMAC_CFG &= (~AT91C_EMAC_FD);
		if( (stat25 & 1) != 0) {
   			AT91F_DBGU_Printk("100Mb Half dup\n");
    		pEmac->EMAC_CFG |= AT91C_EMAC_SPD ;
		} else {
   			AT91F_DBGU_Printk("10Mb Half dup\n");
    		pEmac->EMAC_CFG &= (~AT91C_EMAC_SPD);
		}
	}

	if ((stat1 & (1 << 2)) != 0) {
    	AT91F_DBGU_Printk("Link Up\n"); 
	} else { 
    	AT91F_DBGU_Printk("Link Down\n");
	}
#else 											/* RTL8201BL */
  unsigned short stat1, stat2; 
  
	AT91F_MII_ReadPhy(pEmac, MII_STS_REG, &stat1);
	AT91F_MII_ReadPhy(pEmac, MII_STS_REG, &stat1);

	AT91F_MII_ReadPhy(pEmac, MII_STS2_REG, &stat2);
    
    //set MII for 100BaseTX and Full Duplex
	if ((stat1 & 0x4000) && (stat2 & 0x8000) )
    	pEmac->EMAC_CFG |= (AT91C_EMAC_SPD| AT91C_EMAC_FD);    
    //set MII for 10BaseT and Full Duplex 
	else if ((stat1 & 0x1000) && (stat2 & 0x2000))
    	pEmac->EMAC_CFG = (pEmac->EMAC_CFG & AT91C_EMAC_SPD) | AT91C_EMAC_FD;
    //set MII for 100BaseTX and Half Duplex 
	else if ((stat1 & 0x2000) && (stat2 & 0x4000))
    	pEmac->EMAC_CFG = (pEmac->EMAC_CFG | AT91C_EMAC_SPD ) & ~AT91C_EMAC_FD;
    //set MII for 10BaseT and Half Duplex 
	else if ((stat1 & 0x0800) && (stat2 & 0x1000))
    	pEmac->EMAC_CFG = pEmac->EMAC_CFG & ~AT91C_EMAC_SPD & ~AT91C_EMAC_FD;


	if(AT91F_MII_ReadPhy(pEmac, MII_STS2_REG, & stat2) == 0) {
    	AT91F_DBGU_Printk("PHY status2 register read error\n");
	} else {
    	AT91F_DBGU_Printw("PHY status2", stat2, "\n");
	}
#endif 											
	return 0;
}

//*----------------------------------------------------------------------------
//* \fn    AT91F_MDIO_StartupPhy
//* \brief This function start the phy
//*----------------------------------------------------------------------------
int AT91F_MDIO_StartupPhy(AT91PS_EMAC pEmac)     // \arg Pointer to AT91PS_EMAC service
{
	int ret;
	if(pEmac->EMAC_SR & AT91C_EMAC_LINK)
		return 1;

	pEmac->EMAC_CTL |= AT91C_EMAC_MPE;		// Management Port Enable 

	ret = MII_GetLinkSpeed(pEmac);

	pEmac->EMAC_CTL &= ~AT91C_EMAC_MPE;		// Magement Port Disable
  
	return 0;
}


//*----------------------------------------------------------------------------
//* \fn    AT91F_EMACInit
//* \brief This function initialise the ethernet
//*----------------------------------------------------------------------------
int AT91F_EMACInit( // \return Status ( Success = 0)
		AT91PS_EMAC pEmac,     // \arg Pointer to AT91PS_EMAC service
		unsigned int pTdList)
{
	int status;
	int i;

	AT91F_DBGU_Printk("\n\rMAC ");
	for(i = 0; i < 6; ++i) {
		if(i != 0) {
			AT91F_DBGU_Printk(":");
		}
		AT91F_DBGU_Printl(OurEmacAddr[i]);
	}

	AT91F_DBGU_Printk("\n\rIP ");
	for(i = 0; i < 4; ++i) {
		if(i != 0) {
			AT91F_DBGU_Printk(".");
		}
		AT91F_DBGU_Printl(OurIpAddr[i]);
	}
	AT91F_DBGU_Printk("\n\r");

    AT91F_DBGU_Printk("EMAC Init PMC\n");
	AT91F_EMAC_CfgPMC();

    AT91F_DBGU_Printk("EMAC Init PIO\n");
	AT91F_EMAC_CfgPIO();

#ifndef PHY_RMII
/* MII interface */
	AT91F_PIO_CfgPeriph(
		AT91C_BASE_PIOA, // PIO controller base address
		((unsigned int) AT91C_PA14_ERXER   ) | 	 
		((unsigned int) AT91C_PA12_ERX0    ) |	 
		((unsigned int) AT91C_PA13_ERX1    ) |	 
		((unsigned int) AT91C_PA8_ETXEN   ) |	 
		((unsigned int) AT91C_PA16_EMDIO   ) |	 
		((unsigned int) AT91C_PA9_ETX0    ) |	 
		((unsigned int) AT91C_PA10_ETX1    ) |	 
		((unsigned int) AT91C_PA11_ECRS_ECRSDV) |
		((unsigned int) AT91C_PA15_EMDC    ) |	 
		((unsigned int) AT91C_PA7_ETXCK_EREFCK), // Peripheral A
		0); // Peripheral B

	AT91F_PIO_CfgPeriph(
		AT91C_BASE_PIOB, // PIO controller base address
		0,
	   (((unsigned int) AT91C_PB12_ETX2 ) |
		((unsigned int) AT91C_PB13_ETX3 ) |
		((unsigned int) AT91C_PB15_ERX2 ) |
		((unsigned int) AT91C_PB16_ERX3 ) |
		((unsigned int) AT91C_PB17_ERXDV) |
		((unsigned int) AT91C_PB18_ECOL ) |
		((unsigned int) AT91C_PB19_ERXCK))
	);
#endif 														/* PHY_RMII */

	/* clear EMAC CFG register */
	pEmac->EMAC_CFG = 0;

	status = AT91F_MDIO_StartupPhy(pEmac);
	if ( status )
		return status;

	// the sequence write EMAC_SA1L and write EMAC_SA1H must be respected
	pEmac->EMAC_SA1L = 
		((int)OurEmacAddr[2] << 24) | 
		((int)OurEmacAddr[3] << 16) |
		((int)OurEmacAddr[4] <<  8) | 
		      OurEmacAddr[5];
	pEmac->EMAC_SA1H = ((int)OurEmacAddr[0] << 8) | OurEmacAddr[1];

	pEmac->EMAC_RBQP = pTdList;
	pEmac->EMAC_RSR  &= ~(AT91C_EMAC_OVR | AT91C_EMAC_REC | AT91C_EMAC_BNA);

	pEmac->EMAC_HSL   = 0;
	pEmac->EMAC_HSH   = 0;
	pEmac->EMAC_CFG  |=   AT91C_EMAC_CAF;		/* copy all frames */
	pEmac->EMAC_CFG  &= (~AT91C_EMAC_NBC);
	pEmac->EMAC_CFG  &= ~(AT91C_EMAC_CLK);		/* MDC divider 8 */
#ifdef PHY_RMII
	/* reduced MII interface */
	pEmac->EMAC_CFG  |=   AT91C_EMAC_RMII;		/* set RMII flag  */
#else 
	/* MII interface */
	pEmac->EMAC_CFG  &= (~AT91C_EMAC_RMII);		/* clear RMII flag */
#endif 														/* PHY_RMII */
	/* Tx enable & Rx enable */
	pEmac->EMAC_CTL  |= (AT91C_EMAC_TE | AT91C_EMAC_RE);

 	return 0;
}

//*----------------------------------------------------------------------------
//* \fn    AT91F_EmacEntry
//* \brief Initialise Emac to receive packets
//*----------------------------------------------------------------------------
int AT91F_EmacEntry(void)
{
	unsigned int i;
	char *pRxPacket;
	
	pRxPacket = (char *)RxPacket;

	// Initialise tdList descriptor. This descriptor must be WORD aligned
	for (i = 0; i < NB_ETH_RX_PACKETS; ++i) 
	{
		tdList[i].addr = ((unsigned int) (pRxPacket + (i * ETH_PACKET_SIZE))) & 0xFFFFFFFC;
		tdList[i].size = 0;
	}
	
	// Set the WRAP bit at the end of the list descriptor
	tdList[NB_ETH_RX_PACKETS-1].addr |= 0x02;
	
	return(AT91F_EMACInit(AT91C_BASE_EMAC, (unsigned int) tdList));
}	



//*----------------------------------------------------------------------------
//* \fn    AT91F_IcmpChksum
//* \brief Process ICMP Checksum...
//*----------------------------------------------------------------------------
unsigned short AT91F_IcmpChksum(unsigned short *p, int len) 
{
	int i, t;
	
	for (i=0,t=0; i<len; i++, p++)
		t += SWAP16(*p);
	t = (t & 0xffff) + (t >> 16);
	return (~t);									
}



//*----------------------------------------------------------------------------
//* \fn    AT91F_ProcessEmacPacket
//* \brief Process ARP and ICMP packets...
//*----------------------------------------------------------------------------
int AT91F_ProcessEmacPacket(AT91PS_IPheader pHeader)
{
	unsigned short *pFrameType, *pArpOp;
	unsigned int i, icmp_len;
	char *pData;
	int status = AT91C_NO_IPPACKET;
	
	AT91PS_IPheader pIpHeader;
	AT91PS_IcmpEchoHdr pIcmpEcho;
	int process = 0;
	

	// Receive one packet		
	process = 0;	
	status = 0;
	for (i = 0; i < NB_ETH_RX_PACKETS; ++i) {				
		if(tdList[i].addr & 0x1) {
			if (*AT91C_EMAC_RSR & AT91C_EMAC_REC)
				(*AT91C_EMAC_RSR) |= AT91C_EMAC_REC;
			process = 1;	
			break;				
		}
	}


	if (!process) {
		if (*AT91C_EMAC_RSR & AT91C_EMAC_BNA) {
			(*AT91C_EMAC_CTL) &= (~AT91C_EMAC_RE);
			(*AT91C_EMAC_CTL) |=   AT91C_EMAC_RE;
		}		
		return AT91C_NO_IPPACKET;
	}						

	process = i;
		
	// Process this packet
	pFrameType = (unsigned short *) ((tdList[i].addr & 0xFFFFFFFC) + 12);
	pData      = (char *)(tdList[i].addr & 0xFFFFFFFC);

	switch (SWAP16(*pFrameType)) {
	case PROT_ARP: // ARP Packet format

#if AT91C_DISPLAY_ALL_IPHEADER
		status = AT91C_IPPACKET;
#endif		

		pArpOp = (unsigned short *) (pData + 20);

		if (SWAP16(*pArpOp) == ARP_REQUEST) {
		       // ARP REPLY operation
			*pArpOp =  SWAP16(ARP_REPLY);				
			// Fill the dest address and src address
			for (i = 0; i <6; i++) {
				// swap ethernet dest address and ethernet src address
				pData[i] = pData[i+6];
				pData[i+6] = OurEmacAddr[i];
				// swap sender ethernet address and target ethernet address
				pData[i+22] = OurEmacAddr[i];
				pData[i+32] = pData[i+6];
			}									
				// swap sender IP address and target IP address
			for (i = 0; i<4; i++) {				
				pData[i+38] = pData[i+28];
				pData[i+28] = OurIpAddr[i];
			}
				
			if ( (*AT91C_EMAC_TSR & AT91C_EMAC_BNQ) == 0) {
				/* Tx busy */
				if (*AT91C_EMAC_RSR & AT91C_EMAC_BNA) {
					(*AT91C_EMAC_CTL) &= (~AT91C_EMAC_RE);
					(*AT91C_EMAC_CTL) |=   AT91C_EMAC_RE;
				}		
				tdList[process].addr &= ~0x01;
				return AT91C_NO_IPPACKET;
			}
		  	*AT91C_EMAC_TSR |= AT91C_EMAC_COMP;
			*AT91C_EMAC_TAR = (unsigned int)pData;
 			*AT91C_EMAC_TCR = 0x40;
		}			
	break; // case PROT_ARP
		
	case PROT_IP:	// IP protocol frame
		pIpHeader = (AT91PS_IPheader)(pData + 14);			
		pIcmpEcho = (AT91PS_IcmpEchoHdr)((char *)pIpHeader + 20);
		{ // memcpy(pHeader, pIpHeader,sizeof(AT91S_IPheader));
		    int sz = sizeof(AT91S_IPheader);
		    char *src = (char *) pIpHeader;
		    char *dst = (char *) pHeader;
		    while(sz-- > 0) {
			*dst++ = *src++; 
		    }

		}	
#if AT91C_DISPLAY_ALL_IPHEADER
		status = AT91C_IPPACKET;
#endif		
		
		switch(pIpHeader->ip_p) {
		case PROT_ICMP:		
			// set the status variable to display in main.c only ICMP packets
			status = AT91C_IPPACKET;
				
			// if ICMP_ECHO_REQUEST ==> resp = ICMP_ECHO_REPLY
			if(pIcmpEcho->type == ICMP_ECHO_REQUEST) {
				pIcmpEcho->type = ICMP_ECHO_REPLY;
				pIcmpEcho->code = 0;
				pIcmpEcho->cksum = 0;

				// Checksum of the ICMP Message				
				icmp_len = (SWAP16(pIpHeader->ip_len) - 20)/(sizeof(unsigned short));
			    pIcmpEcho->cksum = SWAP16(AT91F_IcmpChksum((unsigned short *) pIcmpEcho, icmp_len));
					
				// Swap IP Dest address and IP Source address
				for(i = 0; i <4; i++) {
					pIpHeader->ip_dst[i] = pIpHeader->ip_src[i];
					pIpHeader->ip_src[i] = OurIpAddr[i];
				}

				// Swap Eth Dest address and Eth Source address
				for(i = 0; i <6; i++) {
					pData[i] = pData[i + 6];
					pData[i+6] = OurEmacAddr[i];
				}
					// send the echo_reply
			  	*AT91C_EMAC_TSR |= AT91C_EMAC_COMP;
				*AT91C_EMAC_TAR = (unsigned int)pData;
				*AT91C_EMAC_TCR = SWAP16(pIpHeader->ip_len)+ 14;
			}
		break; // case PROT_ICMP

		default:
			break;
		}
	break; // case PROT_IP
						
	default:
		break;
	}// switch (SWAP16(*pFrameType))
	// process the packet
	tdList[process].addr &= ~0x01;
	if (*AT91C_EMAC_RSR & AT91C_EMAC_BNA) {
		(*AT91C_EMAC_CTL) &= (~AT91C_EMAC_RE);
		(*AT91C_EMAC_CTL) |=   AT91C_EMAC_RE;
	}		
	return status;
}